/* Copyright 1998-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

#define BOOL int

#include <iostream>
#include <string.h>
using std::cerr;
using std::size_t;

#include "qubx_util.h"
#include "max_ll_util.h"
#include "ublas_matrixutil.h"
#include "asmffi.h"

extern "C" void QUBFAST_API qubx_QtoPe(double *Q, double *Pe, int n) {
  matrix<double> mQ(n, n), mPe(1, n);
  matrix_from_array(mQ, Q, n, n);
  QtoPe(mQ, mPe);
  matrix_to_array(mPe, Pe, 1, n);
}

extern "C" void QUBFAST_API qubx_QtoQe(double *Q, int n, int *clazz, double tdead, double *Qe) {
  matrix<double> mQ(n, n), mQe(n, n);
  matrix_from_array(mQ, Q, n, n);
  QtoQe(mQ, clazz, tdead, mQe);
  matrix_to_array(mQe, Qe, n, n);
}

extern "C" void QUBFAST_API qubx_expm(int n, double *A, double *Q, double t) {
  matrix<double> mQ(n, n), mA(n, n);
  matrix_from_array(mQ, Q, n, n);
  expm(mQ, t, mA);
  matrix_to_array(mA, A, n, n);
}

extern "C" void QUBFAST_API qubx_QtoQflux(int n, double *A, double *Q) {
  matrix<double> mQ(n, n), mPe(1, n), mPd(n, n), mA(n, n);
  matrix_from_array(mQ, Q, n, n);
  QtoPe(mQ, mPe);
  mPd.clear();
  for (int i=0; i<n; ++i)
    mPd(i,i) = mPe(0,i);
  noalias(mA) = prod(mPd, mQ);
  matrix_to_array(mA, A, n, n);
}

extern "C" void QUBFAST_API qubx_MeanFirstPassage(int n, double *A, double *Q, double sampling) {
  matrix<double> mQ(n, n), mPe(1, n), mA(n, n);
  matrix_from_array(mQ, Q, n, n);
  QtoPe(mQ, mPe);
  expm(mQ, sampling, mA);
  identity_matrix<double> I(n,n);
  scalar_matrix<double> U(n, 1, 1.0);
  matrix<double> Z(n, n), Zi(n, n);
  noalias(Zi) = I - mA + prod(U,mPe);
  invm(Zi, Z);
  for (int i=0; i<n; ++i)
    for (int j=0; j<n; ++j)
      A[i*n+j] = sampling * (Z(j,j) - Z(i,j)) / mPe(0,j);
}

extern "C" void QUBFAST_API qubx_FirstLatencies(double* A, double* pEq, double dt, int n, double* FL)
{
	// from http://www.icsu-scope.org/downloadpubs/scope34/contents.html
	//              Z = inv(I - A + col(1)*row(pEq))
	//       FL[i][j] = (Z[j][j] - Z[i][j]) / pEq[j]

	fq::matrix<double> M(n, n);
	for (int i=0; i<n; ++i)
		for (int j=0; j<n; ++j)
			M[i][j] = ((i==j)?1.0:0.0) - A[i*n+j] + pEq[j];
	
	fq::matrix<double> Z(n, n);
	sv_invm(M, n, Z);

	for (int i=0; i<n; ++i)
		for (int j=0; j<n; ++j)
			FL[i*n+j] = dt * (Z[j][j] - Z[i][j]) / pEq[j];
}

// returns na == number of states in class c == number of components
// outputs:
//   pdf[nbin*(1+na)]; first sum, then each component
//   tau[na], amp[na]
extern "C" int QUBFAST_API qubx_mil_pdf(int n, double *Q, int *clazz, int a, double tdMS, int nbin, float *bins, float *pdf, float *tau, float *amp) {
  matrix<double> mQ(n, n), mQe(n, n);
  matrix_from_array(mQ, Q, n, n);
  QtoQe(mQ, clazz, 1e-3*tdMS, mQe);

  std::vector<int> ia, za, ai(n), zi(n);
  for (int i=0; i<n; ++i) {
    if ( clazz[i] == a ) {
      ai[i] = (int) ia.size();
      ia.push_back(i);
    } else {
      zi[i] = (int) za.size();
      za.push_back(i);
    }
  }
  int na = (int) ia.size();
  int nz = n - na;
  matrix<double> Qaa(na,na), Qaz(na,nz), Qza(nz,na), Qzz(nz,nz), aPe(1, na), spectre(na,na);
  matrix<double> tmpAA(na, na), tmpAZ(na, nz), tmpZZ;
  std::vector<double> W(na), Wimag(na);
  matrix<double> V, Vi;
  
  for (int i=0; i<n; ++i) {
    for (int j=0; j<n; ++j) {
      if ( clazz[i] == a ) {
        if ( clazz[j] == a )
          Qaa(ai[i],ai[j]) = mQe(i,j);
        else
          Qaz(ai[i],zi[j]) = mQe(i,j);
      } else {
        if ( clazz[j] == a )
          Qza(zi[i],ai[j]) = mQe(i,j);
        else
          Qzz(zi[i],zi[j]) = mQe(i,j);
      }
    }
  }

  matrix<double> QaaiQazQzziQzaT(na, na);
  invm(Qaa, tmpAA);
  invm(Qzz, tmpZZ);
  noalias(tmpAZ) = prod(tmpAA, Qaz);
  tmpAZ = prod(tmpAZ, tmpZZ);
  noalias(tmpAA) = prod(tmpAZ, Qza);
  noalias(QaaiQazQzziQzaT) = trans(tmpAA);
  eigen_vals_vecs(QaaiQazQzziQzaT, &W[0], &Wimag[0], V, Vi, console);
  double vUnit = fabs(W[0]-1.0);
  double iUnit = 0;
  for (int i=1; i<na; ++i) {
    double absV = fabs(W[i]-1.0);
    if ( absV < vUnit ) {
      vUnit = absV;
      iUnit = i;
    }
  }
  double sumPe = 0.0;
  for (int i=0; i<na; ++i) {
    vUnit = V(i, iUnit);
    aPe(0,i) = vUnit;
    sumPe += vUnit;
  }
  if ( sumPe )
    for (int i=0; i<na; ++i)
      aPe(0,i) /= sumPe;
  
  eigen_vals_vecs(Qaa, &W[0], &Wimag[0], V, Vi, console);

  memset(pdf, 0, na*nbin*sizeof(float));
  for (int i=0; i<na; ++i) {
    float *component = pdf + (i+1)*nbin;
    for (int j=0; j<na; ++j) 
      for (int k=0; k<na; ++k)
        spectre(j,k) = V(j,i) * Vi(i,k);

    matrix<double> tmp1(1, na), tmp2(1, nz);
    noalias(tmp1) = prod(aPe, spectre);
    noalias(tmp2) = prod(tmp1, Qaz);
    double coeff = 0.0;
    for (int j=0; j<nz; ++j) {
      coeff += tmp2(0,j);
    }
    coeff /= W[i];
    double itau = W[i];
    double last = bins[0] * bins[0] / bins[1];
    for (int b=0; b<nbin; ++b) {
      double next = bins[b];
      float bar = (float) (coeff * (exp(itau*(next-tdMS)*1e-3) - exp(itau*(last-tdMS)*1e-3)));
      component[b] = bar;
      pdf[b] += bar;
      last = next;
    }
    tau[i] = -1.0f/(float)itau;
    amp[i] = (float) -coeff;
  }
  std::vector<int> sorted(na);
  for (int i=0; i<na; ++i)
    sorted[i] = i;
  sort(sorted.begin(), sorted.end(), [tau](const int i1, const int i2) -> bool {
      return tau[i1] < tau[i2];
    });
  /// reorder
  return na;
}
